/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _DIALOG_H_
#define _DIALOG_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 


#include <qcheckbox.h>
#include <qcombobox.h>
#include <qdialog.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qradiobutton.h>
#include <qspinbox.h> 
#include <qstring.h> 

#include "hexbuffer.h"
#include "hexvalidator.h"

const int repYes      = QDialog::Accepted; 
const int repNo       = 11;
const int repAll      = 12;
const int repClose    = QDialog::Rejected; 
const int repNext     = 13;
const int repPrevious = 14;
const int repNewKey   = 15;

const uint Replace_All     = 1;
const uint Replace_AllInit = 2;
const uint Replace_First   = 3;
const uint Replace_Next    = 4;
const uint Replace_Ignore  = 5;
const uint Find_First      = 6;
const uint Find_Next       = 7;


class CGotoDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CGotoDialog( QWidget *parent, const char *name = 0 );
    ~CGotoDialog( void );
    void defaultFocus( void );

  signals:
    void gotoOffset( uint offset, uint bit, bool fromCursor, bool forward );

  private slots:
    void startGoto( void );

  private:
    QComboBox *mComboBox;
    QCheckBox *mCheckBackward;
    QCheckBox *mCheckFromCursor;  
};


class CFindDialog : public QDialog 
{
  Q_OBJECT

  public:
    enum EOperation
    {
      find_Again = 0,
      find_Next,
      find_Previous,
    };

  public:
    CFindDialog( QWidget *parent, const char *name = 0 );
    ~CFindDialog( void );
 
    void defaultFocus( void );
    const char *data( void );

  signals: 
    void findData( SSearchControl &sc, uint mode, bool navigator );

  public slots:
    void findAgain( EOperation operation );

  private slots:
    void startFind( void );
    void selectorChanged( int index );
    void inputChanged( const char *text );

  private:
    uint size( void );

  private:
    QComboBox *mSelector;
    QLineEdit *mInput;
    QCheckBox *mCheckBackward;
    QCheckBox *mCheckFromCursor;
    QCheckBox *mCheckInSelection;
    QCheckBox *mCheckUseNavigator;

    QString    mFindString[4];
    QByteArray mFindData;
    CHexValidator *mFindValidator;
};



class CFindNavigatorDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CFindNavigatorDialog( QWidget *parent, const char *name = 0 );
    ~CFindNavigatorDialog( void );
    void defineData( SSearchControl &sc );
  
  signals:
    void findData( SSearchControl &sc, uint mode, bool navigator );
    void makeKey( void );

  private slots:
    void next( void );
    void previous( void );
    void newKey( void );
    void close( void );

  private:
    void done( int returnCode );

  private:
    QLineEdit *mKey;
    SSearchControl mSearchControl;
};


class CReplaceDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CReplaceDialog( QWidget *parent, const char *name = 0 );
    ~CReplaceDialog( void );
    void defaultFocus( void );

  signals:
    void replaceData( SSearchControl &sc, uint mode );

  private slots:
    void startReplace( void );
    void findSelectorChanged( int index );
    void findInputChanged( const char *text );
    void replaceSelectorChanged( int index );
    void replaceInputChanged( const char *text );

  private:
    const char *findData( void );
    uint findSize( void );
    const char *replaceData( void );
    uint replaceSize( void );

  private:
    QComboBox *mFindSelector;
    QComboBox *mReplaceSelector;
    QLineEdit *mFindInput;
    QLineEdit *mReplaceInput;

    QCheckBox *mCheckBackward;
    QCheckBox *mCheckFromCursor;
    QCheckBox *mCheckInSelection;
    QCheckBox *mCheckPrompt;

    QString    mFindString[4];
    QString    mReplaceString[4];
    QByteArray mFindData;
    QByteArray mReplaceData;
    CHexValidator *mFindValidator;
    CHexValidator *mReplaceValidator;
};



class CReplacePromptDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CReplacePromptDialog( QWidget *parent, const char *name = 0 );
    ~CReplacePromptDialog( void );
    void defineData( SSearchControl &sc );
  
  signals:
    void replaceData( SSearchControl &sc, uint mode );

  private slots:
    void no( void );
    void yes( void );
    void all( void );
    void close( void );

  private:
    void done( int returnCode );

  private:
    SSearchControl mSearchControl;
};





class CInsertDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CInsertDialog( QWidget *parent, const char *name = 0 );
    ~CInsertDialog( void );
    void defaultFocus( void );

  signals:
    void execute( SInsertData &id );

  private slots:
    void startInsert( void );
    void cursorCheck( void );
    void patternSelectorChanged( int index );
    void patternInputChanged( const char *text );

  private:
    const char *pattern( void );
    uint patternSize( void );

  private:
    QSpinBox *mSizeBox;
    QLabel *mOffsetLabel;
    QComboBox *mPatternSelector;
    QLineEdit *mPatternInput;
    QLineEdit *mOffsetInput;
    QCheckBox *mCheckPattern;
    QCheckBox *mCheckOnCursor;

    QString    mPatternString[4];
    QByteArray mPatternData;
    CHexValidator *mPatternValidator;
};





class CExportDialog : public QDialog 
{
  Q_OBJECT
  
  public:
    CExportDialog( QWidget *parent, const char *name = 0 );
    ~CExportDialog( void );
    void defaultFocus( void );

  signals:
    void execute( SExportData &ed );

  private slots:
    void startExport( void );
    void browserClicked( void );
    void radioClicked( int id );
    void checkStdout( void );

  private:
    QLabel *mDestinationLabel;
    QLineEdit *mDestinationInput;
    QRadioButton *mRadioEverything;
    QRadioButton *mRadioSelectedArea;
    QRadioButton *mRadioSpecify;
    QLabel *mFromLabel;
    QLabel *mToLabel;
    QLineEdit *mFromInput;
    QLineEdit *mToInput;
    QCheckBox *mCheckPS;
    QCheckBox *mCheckStdout;
};

void centerDialog( QWidget *widget, QWidget *centerParent );
void centerDialogBottom( QWidget *widget, QWidget *centerParent );

#endif





