/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _HEX_EDITOR_WIDGET_H_
#define _HEX_EDITOR_WIDGET_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 

#include <qprinter.h> 

#include <kapp.h>
#include <kfm.h>

#include "dialog.h"
#include "hexbuffer.h"
#include "hexviewwidget.h"
#include "optiondialog.h"
#include "printerdialog.h"
#include "progress.h"


class CHexEditorWidget : public QWidget
{
  Q_OBJECT

  public:
    enum EProgressMode
    {
      pg_read = 0,
      pg_write,
      pg_insert,
      pg_print,
      pg_encode
    };

  public:
    CHexEditorWidget( QWidget *parent = 0, const char *name = 0 );
    ~CHexEditorWidget( void );
    void initialize( void );
    void writeConfiguration( KConfig &config );
    void readConfiguration( KConfig &config );

    void open( const QString &url );
    bool selectDocument( const QString &url );
    bool modified( void );

    inline int defaultTextWidth( void );
    inline CHexViewWidget *view( void );
    inline SDisplayLayout &layout( void );
    inline SDisplayLine &line( void );
    inline SDisplayInputMode &inputMode( void );
    inline SDisplayMisc::EOpenFile openFile( void );

  signals:
    void errorLoadFile( const QString &url );
    void setProgress( int percent );
    void enableProgressText( bool state );
    void setProgressText( const QString &msg );
    void operationChanged( bool state );

  public slots:
    void fontChanged( void );
    void paletteChanged( void );
    void layoutChanged( const SDisplayLayout &layout );
    void inputModeChanged( const SDisplayInputMode &input );

    void setLineSize(const SDisplayLine &line );
    void setLayout( const SDisplayLayout &layout );
    void setCursor( const SDisplayCursor &cursor );
    void setColor( const SDisplayColor &color );
    void setFont( const SDisplayFont &font );
    void setMisc( const SDisplayMisc &misc );

    void newFile( void );
    void open( void );
    void insertFile( void );
    void stop( void );
    bool close( void );
    bool closeAll( void );
    void save( void );
    void saveAs( void );
    void reload( void );
    void print( void );
    void printData( QPrinter & );
    void export( void );
    void encode( CConversion::EMode mode );
    void undo( void );
    void redo( void );
    void toggleWriteProtection( void );
    void defaultWriteProtection( void );
    void benchmark( void );
    void copy( void );
    void copyPrimaryFormatted( void );
    void copyAllFormatted( void );
    void paste( void );
    void pasteNewFile( void );
    void cut( void );
    void selectAll( void );
    void unselect( void );
    void addBookmark( void );
    void removeBookmark( bool all );
    void replaceBookmark( void );
    void gotoBookmark( uint position );
    void gotoOffset( void );
    void find( void );
    void findData( SSearchControl &sc, uint mode, bool navigator );
    void replaceData( SSearchControl &sc, uint mode );
    void replacePrompt( SSearchControl &sc );
    void replaceResult( SSearchControl &sc );
    void findAgain( void );
    void findNext( void );
    void findPrevious( void );
    void findNavigator( SSearchControl &sc );
    void replace( void );
    void insertPattern( void );
    void options( void );

    void kfmFinished( void );
    void kfmError( int errCode, const char *errString );

  protected:
    void resizeEvent( QResizeEvent *e );

  private:
    enum action
    { 
      GET, 
      PUT 
    }; 

  private slots:  
    void exportData( SExportData &ed );

  private:
    bool querySave( void );
    int  readURL( const QString &url, bool insert );
    void writeURL( QString &url );
    bool readFile( const QString &diskPath, const QString &url, bool insert );
    bool writeFile( QString &diskPath );
    KFM *CreateKFMConnection( void );
    void DestroyKFMConnection( void );

    CHexBuffer *documentItem( const QString &url );
    bool createBuffer( void );
    void removeBuffer( void );

    bool askWrap( bool fwd, const QString &header );
    bool canFind( bool showError );
    void hideReplacePrompt( void );


    int prepareProgressData( EProgressMode mode );
    static int progressReceiver( void *clientData, float fraction );
    int progressParse( float fraction );

  private:
    KFM      *mKfm;
    QString  mKfmURL;
    QString  mKfmTMP;
    action   mKfmAction;
    QPrinter *mPrinter;

    bool mPrintingOn;
    uint mUntitledCount;
    QList<CHexBuffer> mDocumentList;
    SDisplayState mDisplayState;

    CProgress     mProgressData;
    EProgressMode mProgressMode;
    bool          mProgressBusy;
    bool          mProgressStop;

    CHexViewWidget *mHexView;
    CGotoDialog    *mGotoDialog;
    CFindDialog    *mFindDialog;
    CFindNavigatorDialog *mFindNavigatorDialog;
    CReplaceDialog *mReplaceDialog;
    CReplacePromptDialog *mReplacePromptDialog;
    CInsertDialog  *mInsertDialog;
    CExportDialog  *mExportDialog;
    COptionDialog  *mOptionDialog;
    CPrinterDialog *mPrintDialog;
};

inline int CHexEditorWidget::defaultTextWidth( void )
{
  return( mHexView->lineWidth() + mHexView->scrollBarWidth() );
}

inline CHexViewWidget *CHexEditorWidget::view( void )
{
  return( mHexView );
}

inline SDisplayLayout &CHexEditorWidget::layout( void )
{
  return( mDisplayState.layout );
}

inline SDisplayLine &CHexEditorWidget::line( void )
{
  return( mDisplayState.line );
}

inline SDisplayInputMode &CHexEditorWidget::inputMode( void )
{
  return( mDisplayState.input );
}

inline SDisplayMisc::EOpenFile CHexEditorWidget::openFile( void )
{
  return( mDisplayState.misc.openFile );
}


#endif



