/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *   This file is based on the work by Martynas Kunigelis (KProgress)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "knewprogress.h"

#include <qpainter.h>
#include <qstring.h>
#include <qpixmap.h>
#include <kapp.h>

KNewProgress::KNewProgress( QWidget *parent, const char *name )
  : QFrame(parent, name),
    QRangeControl(0, 100, 1, 10, 0),
    orient( Horizontal )
{
  initialize();
}

KNewProgress::KNewProgress( Orientation orientation, QWidget *parent, 
			    const char *name )
  : QFrame(parent, name),
    QRangeControl(0, 100, 1, 10, 0),
    orient( orientation )
{
  initialize();
}

KNewProgress::KNewProgress( int minValue, int maxValue, int value, 
			    Orientation orientation, QWidget *parent, 
			    const char *name )
  : QFrame(parent, name),
    QRangeControl(minValue, maxValue, 1, 10, value),
    orient(orientation)
{
  initialize();
}

KNewProgress::~KNewProgress()
{
  delete bar_pixmap;
}

void KNewProgress::advance( int offset )
{
  setValue(value() + offset);
}

void KNewProgress::initialize()
{
  //setFrameStyle(QFrame::Panel | QFrame::Sunken);
  //setLineWidth(2);
  //setMidLineWidth(2);
  bar_pixmap = 0;
  bar_style = Solid;
  bar_color = kapp->selectColor;
  bar_text_color = kapp->selectTextColor;
  text_color = kapp->textColor;
  setBackgroundColor( kapp->backgroundColor );
  setFont(QFont("helvetica", 12, QFont::Bold));
  text_enabled = TRUE;
  adjustStyle();
}


void KNewProgress::setBarPixmap(const QPixmap &pixmap)
{
  if(pixmap.isNull())
    return;
  if(bar_pixmap)
    delete bar_pixmap;
		
  bar_pixmap = new QPixmap(pixmap);
}

void KNewProgress::setBarColor(const QColor &color)
{
  bar_color = color;
  if( bar_pixmap != 0 ) 
  {
    delete bar_pixmap; bar_pixmap = 0;
  }
}

void KNewProgress::setBarStyle(BarStyle style)
{
  if( bar_style != style ) 
  {
    bar_style = style;
    update();
  }
}

void KNewProgress::setOrientation(Orientation orientation)
{
  if( orient != orientation) 
  {
    orient = orientation;
    update();
  }
}
 
void KNewProgress::setValue(int value)
{
  QRangeControl::setValue(value);
}

void KNewProgress::setTextEnabled(bool enable)
{
  if( text_enabled != enable )
  {
    text_enabled = enable;
    update();
  }
}

void KNewProgress::setText( const QString &msg )
{
  mMsg.duplicate( msg ); 
  if( text_enabled == TRUE )
  {
    update();
  }
}




const QColor & KNewProgress::barColor() const
{
  return bar_color;
}

const QPixmap * KNewProgress::barPixmap() const
{
  return bar_pixmap;
}

bool KNewProgress::textEnabled() const
{
  return text_enabled;
}

QSize KNewProgress::sizeHint() const
{
  QSize s( size() );

  if(orientation() == KNewProgress::Vertical) 
  {
    s.setWidth(24);
  } 
  else 
  {
    s.setHeight(24);
  }

  return s;
}


KNewProgress::Orientation KNewProgress::orientation() const
{
  return orient;
}

KNewProgress::BarStyle KNewProgress::barStyle() const
{
  return bar_style;
}

int KNewProgress::recalcValue(int range)
{
  int abs_value = value() - minValue();
  int abs_range = maxValue() - minValue();
  return range * abs_value / abs_range;
}

void KNewProgress::valueChange()
{
  repaint(contentsRect(), FALSE);
  emit percentageChanged(recalcValue(100));
}

void KNewProgress::rangeChange()
{
  repaint(contentsRect(), FALSE);
  emit percentageChanged(recalcValue(100));
}

void KNewProgress::styleChange(GUIStyle)
{
  adjustStyle();
}

void KNewProgress::adjustStyle()
{
  switch (style()) 
  {
    case WindowsStyle:
      setFrameStyle(QFrame::WinPanel | QFrame::Sunken);
    break;
    case MotifStyle:
    default:
      setFrameStyle(QFrame::Panel | QFrame::Sunken);
      setLineWidth( 2 );
    break;
  }
  update();
}

void KNewProgress::paletteChange( const QPalette & )
{
  bar_color = kapp->selectColor;
  bar_text_color = kapp->selectTextColor;
  text_color = kapp->textColor;
  setBackgroundColor( kapp->backgroundColor );
}
		
void KNewProgress::drawText( QPainter *p )
{
  QRect r(contentsRect());
  QString s;
	
  if( mMsg.isEmpty() == TRUE )
  {
    s.sprintf( "%i%%", recalcValue(100) );
  }
  else
  {
    s.sprintf( "%s... %i%%", mMsg.data(), recalcValue(100) );
  }

  p->setPen(text_color);
  p->drawText(r, AlignCenter, s);
  p->setClipRegion( fr );
  p->setPen(bar_text_color);
  p->drawText(r, AlignCenter, s);
}

void KNewProgress::drawContents(QPainter *p)
{
  QRect cr = contentsRect(), er = cr;
  fr = cr;
  QBrush fb(bar_color), eb(backgroundColor());

  if (bar_pixmap)
    fb.setPixmap(*bar_pixmap);

  if (backgroundPixmap())
    eb.setPixmap(*backgroundPixmap());

  switch (bar_style) 
  {
    case Solid:
      if (orient == Horizontal) 
      {
	fr.setWidth(recalcValue(cr.width()));
	er.setLeft(fr.right() + 1);
      } 
      else 
      {
	fr.setTop(cr.bottom() - recalcValue(cr.height()));
	er.setBottom(fr.top() - 1);
      }
				
      p->setBrushOrigin(cr.topLeft());
      p->fillRect(fr, fb);
      p->fillRect(er, eb);

      if( text_enabled )
	drawText(p);
    break;
			
    case Blocked:
      const int margin = 2;
      int max, num, dx, dy;
      if (orient == Horizontal) 
      {
	fr.setHeight(cr.height() - 2 * margin);
	fr.setWidth((int)(0.67 * fr.height()));
	fr.moveTopLeft(QPoint(cr.left() + margin, cr.top() + margin));
	dx = fr.width() + margin;
	dy = 0;
	max = (cr.width() - margin) / (fr.width() + margin) + 1;
	num = recalcValue(max);
      } 
      else 
      {
	fr.setWidth(cr.width() - 2 * margin);
	fr.setHeight((int)(0.67 * fr.width()));
	fr.moveBottomLeft(QPoint(cr.left() + margin, cr.bottom() - margin));
	dx = 0;
	dy = - (fr.height() + margin);
	max = (cr.height() - margin) / (fr.height() + margin) + 1;
	num = recalcValue(max);
      }
      p->setClipRect(cr.x() + margin, cr.y() + margin, 
		     cr.width() - margin, cr.height() - margin);
      for (int i = 0; i < num; i++) 
      {
	p->setBrushOrigin(fr.topLeft());
	p->fillRect(fr, fb);
	fr.moveBy(dx, dy);
      }
      
      if (num != max) 
      {
	if (orient == Horizontal) 
	  er.setLeft(fr.right() + 1);
	else
	  er.setBottom(fr.bottom() + 1);
	if (!er.isNull()) 
	{
	  p->setBrushOrigin(cr.topLeft());
	  p->fillRect(er, eb);
	}
      }  
    break;
  }	
		
}			
