/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <qbttngrp.h>
#include <qfont.h>
#include <qlayout.h>
#include <qlabel.h>

#include <kapp.h>

#include "dialogmargin.h"
#include "optiondialog.h"


COptionDialog::COptionDialog( const char *title, QWidget *parent, 
			      char *name, bool modal )
  : CPageDialog( CPageDialog::TREE,  CPageDialog::HELP | CPageDialog::DEFAULT |
		 CPageDialog::OK | CPageDialog::APPLY | CPageDialog::CANCEL,
		 CPageDialog::OK, title, parent, name, modal )
{
  const char *longTitle;

  longTitle = i18n("Data layout in editor.");
  mFrame[ page_layout ] = addPage( -1, i18n("Layout"), longTitle );
  longTitle = i18n("Cursor behaviour. Only valid for editor.");
  mFrame[ page_cursor ] = addPage( -1, i18n("Cursor"), longTitle );
  longTitle = i18n("Editor colors. System selection color is always used.");
  mFrame[ page_color ]  = addPage( -1, i18n("Colors"), longTitle );
  longTitle = i18n("Font selection. Editor can only use a fixed font.");
  mFrame[ page_font ]   = addPage( -1, i18n("Font"), longTitle );
  longTitle = i18n("Various properties.");
  mFrame[ page_misc ] = addPage( -1, i18n("Miscellaneous"), longTitle );

  SetupLayoutPage();
  SetupCursorPage();
  SetupColorPage();
  SetupFontPage();
  SetupMiscPage();

  connect( this, SIGNAL(defaultClicked()), SLOT(slotDefault()));
  connect( this, SIGNAL(applyClicked()), SLOT(slotApply()));
  
  adjustPageSize();
  adjustSize();
  showPage( page_layout );
}

COptionDialog::~COptionDialog( void )
{
}




void COptionDialog::SetupLayoutPage( void )
{
  const char *text;
  QFrame *page = mFrame[ page_layout ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );
  if( topLayout == 0 ) { return; }

  QFrame *frame = new QFrame( page );
  if( frame == 0 ) { return; }
  topLayout->addWidget( frame, 0, AlignLeft );
  QGridLayout *gbox = new QGridLayout( frame, 13, 2, 0, INNER_SEP );
  if( gbox == 0 ) { return; }

  mLayout.formatCombo  = new QComboBox( false, frame );
  mLayout.formatCombo->setFixedHeight( 
    mLayout.formatCombo->sizeHint().height() );
  mLayout.formatCombo->setMinimumWidth( 
    mLayout.formatCombo->sizeHint().width() );
  const char *items[] = 
  { 
    i18n("Hexadecimal mode"),
    i18n("Octal mode"),
    i18n("Binary mode"),
    i18n("Text only mode"),
    0 
  };
  mLayout.formatCombo->insertStrList( items );
  connect( mLayout.formatCombo, SIGNAL(activated(int)), 
	   SLOT(slotModeSelectorChanged(int)) );
  gbox->addWidget( mLayout.formatCombo, 0, 0 );

  mLayout.lineSizeSpin = new QSpinBox( frame );
  mLayout.lineSizeSpin->setFixedHeight( 
    mLayout.lineSizeSpin->sizeHint().height() );
  mLayout.lineSizeSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.lineSizeSpin->setRange( 1, 10000 );
  connect( mLayout.lineSizeSpin, SIGNAL(valueChanged(int)), 
	   SLOT(slotLineSizeChanged(int) ) );
  gbox->addWidget( mLayout.lineSizeSpin, 1, 1 );

  mLayout.columnSizeSpin = new QSpinBox( frame );
  mLayout.columnSizeSpin->setFixedHeight( 
    mLayout.columnSizeSpin->sizeHint().height() );
  mLayout.columnSizeSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.columnSizeSpin->setRange( 1, 10000 );
  connect( mLayout.columnSizeSpin, SIGNAL(valueChanged(int)), 
	   SLOT(slotColumnSizeChanged(int) ) );
  gbox->addWidget( mLayout.columnSizeSpin, 2, 1 );

  text = i18n("Default l&ine size [bytes]");
  QLabel *label = new QLabel( mLayout.lineSizeSpin, text, frame );
  label->setMinimumWidth( label->sizeHint().width() );
  gbox->addWidget( label, 1, 0 );

  text = i18n("Colu&mn size [bytes]");
  label = new QLabel( mLayout.columnSizeSpin, text, frame );
  label->setMinimumWidth( label->sizeHint().width() );
  gbox->addWidget( label, 2, 0 );

  QFrame *hline = new QFrame( frame );
  hline->setFrameStyle( QFrame::Sunken | QFrame::HLine );
  hline->setFixedHeight( hline->sizeHint().height() );
  gbox->addMultiCellWidget( hline, 3, 3, 0, 1, AlignLeft );

  text = i18n("Line size is &fixed (use scrollbar when required)"); 
  mLayout.lockLineCheck = new QCheckBox( text, frame );
  mLayout.lockLineCheck->setFixedSize( mLayout.lockLineCheck->sizeHint() );
  gbox->addMultiCellWidget( mLayout.lockLineCheck, 4, 4, 0, 1, AlignLeft );

  text = i18n("Loc&k column at end of line (when column size > 1)"); 
  mLayout.lockColumnCheck = new QCheckBox( text, frame );
  mLayout.lockColumnCheck->setFixedSize( mLayout.lockColumnCheck->sizeHint() );
  gbox->addMultiCellWidget( mLayout.lockColumnCheck, 5, 5, 0, 1, AlignLeft );

  hline = new QFrame( frame );
  hline->setFrameStyle( QFrame::Sunken | QFrame::HLine );
  hline->setFixedHeight( hline->sizeHint().height() );
  gbox->addMultiCellWidget( hline, 6, 6, 0, 1, AlignLeft );

  mLayout.leftSepWidthSpin = new QSpinBox( frame );
  mLayout.leftSepWidthSpin->setFixedHeight( 
    mLayout.leftSepWidthSpin->sizeHint().height() );
  mLayout.leftSepWidthSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.leftSepWidthSpin->setRange( 0, 20 );
  gbox->addWidget( mLayout.leftSepWidthSpin, 7, 1 );

  mLayout.rightSepWidthSpin = new QSpinBox( frame );
  mLayout.rightSepWidthSpin->setFixedHeight( 
    mLayout.rightSepWidthSpin->sizeHint().height() );
  mLayout.rightSepWidthSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.rightSepWidthSpin->setRange( 0, 20 );
  gbox->addWidget( mLayout.rightSepWidthSpin, 8, 1 );

  text = i18n("&Left separator width [pixels]");
  mLayout.leftSepLabel = new QLabel( mLayout.leftSepWidthSpin, text, frame );
  mLayout.leftSepLabel->setMinimumWidth( 
    mLayout.leftSepLabel->sizeHint().width() );
  gbox->addWidget( mLayout.leftSepLabel, 7, 0 );

  text = i18n("&Right separator width [pixels]");
  mLayout.rightSepLabel = new QLabel( mLayout.rightSepWidthSpin, text, frame );
  mLayout.rightSepLabel->setMinimumWidth( 
    mLayout.rightSepLabel ->sizeHint().width() );
  gbox->addWidget( mLayout.rightSepLabel, 8, 0 );

  mLayout.separatorSpin = new QSpinBox( frame );
  mLayout.separatorSpin->setFixedHeight( 
    mLayout.separatorSpin->sizeHint().height() );
  mLayout.separatorSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.separatorSpin->setRange( 0, 20 );
  gbox->addWidget( mLayout.separatorSpin, 9, 1 );

  mLayout.edgeSpin = new QSpinBox( frame );
  mLayout.edgeSpin->setFixedHeight( mLayout.edgeSpin->sizeHint().height() );
  mLayout.edgeSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.edgeSpin->setRange( 0, 20 );
  gbox->addWidget( mLayout.edgeSpin, 10, 1 );

  text = i18n("&Separator margin width [pixels]");
  label = new QLabel( mLayout.separatorSpin, text, frame );
  label->setMinimumWidth( label->sizeHint().width() );
  gbox->addWidget( label, 9, 0 );

  text = i18n("&Edge margin width [pixels]");
  label = new QLabel( mLayout.edgeSpin, text, frame );
  label->setMinimumWidth( label->sizeHint().width() );
  gbox->addWidget( label, 10, 0 );
 
  text = i18n("Column separation is e&qual to one character");
  mLayout.columnCheck = new QCheckBox( text, frame );
  mLayout.columnCheck->setFixedSize( mLayout.columnCheck->sizeHint() );
  mLayout.columnCheck->setMinimumSize( mLayout.columnCheck->sizeHint() );
  gbox->addMultiCellWidget( mLayout.columnCheck, 11, 11, 0, 1, AlignLeft );
  connect( mLayout.columnCheck, SIGNAL(toggled(bool)),
	   SLOT( slotColumnSepCheck(bool)));
 
  mLayout.columnSepSpin = new QSpinBox( frame );
  mLayout.columnSepSpin->setFixedHeight( 
    mLayout.columnSepSpin->sizeHint().height() );
  mLayout.columnSepSpin->setMinimumWidth( fontMetrics().width("M") * 10 );
  mLayout.columnSepSpin->setRange( 1, 100 );
 
  text = i18n("Column separa&tion [pixels]");
  mLayout.columnSepLabel = new QLabel( mLayout.columnSepSpin, text, frame );
  mLayout.columnSepLabel->setMinimumWidth( 
    mLayout.columnSepLabel->sizeHint().width() );

  gbox->addWidget( mLayout.columnSepLabel, 12, 0 );
  gbox->addWidget( mLayout.columnSepSpin, 12, 1 );
  gbox->activate();


  QWidget *labelList[3] = 
  { 
    mLayout.lockLineCheck, mLayout.lockColumnCheck, mLayout.columnCheck
  };
  int maxWidth = 0;
  for( int i=0; i<3; i++)
  {
    int w = labelList[i]->sizeHint().width();
    if( w > maxWidth ) { maxWidth = w; }
  }
  frame->setMinimumWidth( maxWidth );


  topLayout->addStretch(10);
  topLayout->activate();
}


void COptionDialog::SetupCursorPage( void )
{
  const char *text;
  QFrame *page = mFrame[ page_cursor ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );
  if( topLayout == 0 ) { return; }


  QGroupBox *group = new QGroupBox( page );
  if( group == 0 ) { return; }
  topLayout->addWidget( group );
  group->setTitle( i18n("Blinking") );
  group->setFrameStyle( QFrame::Box|QFrame::Sunken );
  QVBoxLayout *vbox = new QVBoxLayout( group, 5 );
  vbox->addSpacing( group->fontMetrics().height() );
  text = i18n("Do not b&link");
  mCursor.blinkCheck = new QCheckBox( text, group );
  mCursor.blinkCheck->setFixedHeight(mCursor.blinkCheck->sizeHint().height());
  connect( mCursor.blinkCheck, SIGNAL(toggled(bool)),
	   SLOT( slotBlinkIntervalCheck(bool)));
  vbox->addWidget( mCursor.blinkCheck );
  
  QHBoxLayout *hbox = new QHBoxLayout();
  if( hbox == 0 ) { return; }
  vbox->addLayout( hbox );
  mCursor.blinkSpin = new QSpinBox( group );
  mCursor.blinkSpin->setFixedHeight( mCursor.blinkSpin->sizeHint().height() );
  mCursor.blinkSpin->setRange( 100, 1000 );
  mCursor.blinkSpin->setSteps( 100, 100 );
  mCursor.blinkSpin->setValue( 500 );
  text = i18n("&Blink interval [ms]" );
  mCursor.blinkLabel = new QLabel( mCursor.blinkSpin, text, group );
  hbox->addWidget( mCursor.blinkLabel );
  hbox->addWidget( mCursor.blinkSpin );
  vbox->activate();

  group = new QGroupBox( page );
  if( group == 0 ) { return; }
  topLayout->addWidget( group );
  group->setTitle( i18n("Shape") );
  group->setFrameStyle( QFrame::Box|QFrame::Sunken );
  vbox = new QVBoxLayout( group, 5 );
  vbox->addSpacing( group->fontMetrics().height() );
  text = i18n("Always &use block (rectangular) cursor");
  mCursor.blockCheck = new QCheckBox( text, group );
  mCursor.blockCheck->setFixedHeight( mCursor.blockCheck->sizeHint().height());
  connect( mCursor.blockCheck, SIGNAL(toggled(bool)),
	   SLOT( slotBlockCursorCheck(bool)));
  vbox->addWidget( mCursor.blockCheck );
  text = i18n("Use &thick cursor in insert mode");
  mCursor.thickCheck = new QCheckBox( text, group );
  mCursor.thickCheck->setFixedHeight( mCursor.thickCheck->sizeHint().height());
  vbox->addWidget( mCursor.thickCheck );
  vbox->activate();

  QButtonGroup *buttonGroup = new QButtonGroup( page );
  if( buttonGroup == 0 ) { return; }
  buttonGroup->setTitle( i18n("Cursor behaviour when editor looses focus") );
  topLayout->addWidget( buttonGroup, 0, AlignLeft );
  vbox = new QVBoxLayout( buttonGroup, 5 );
  vbox->addSpacing( buttonGroup->fontMetrics().height() );
  mCursor.stopRadio = new QRadioButton( "radio1", buttonGroup );
  text = i18n("&Stop blinking (if blinking is enabled)");
  mCursor.stopRadio->setText( text );
  mCursor.stopRadio->setMinimumSize( mCursor.stopRadio->sizeHint() );
  vbox->addWidget( mCursor.stopRadio );
  mCursor.hideRadio = new QRadioButton( "radio2", buttonGroup );
  mCursor.hideRadio->setText(i18n("H&ide") );
  mCursor.hideRadio->setMinimumSize( mCursor.hideRadio->sizeHint() );
  vbox->addWidget( mCursor.hideRadio );
  mCursor.nothingRadio = new QRadioButton( "radio3", buttonGroup );
  mCursor.nothingRadio->setText(i18n("Do &nothing") );
  mCursor.nothingRadio->setMinimumSize( mCursor.nothingRadio->sizeHint() );
  vbox->addWidget( mCursor.nothingRadio );
  vbox->activate();

  topLayout->addStretch(10);
  topLayout->activate();
}

void COptionDialog::SetupColorPage( void )
{
  const char *text;
  QFrame *page = mFrame[ page_color ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );
  if( topLayout == 0 ) { return; }

  text = i18n("&Use system colors (as chosen in Control Center)");
  mColor.checkSystem = new QCheckBox( text, page );
  mColor.checkSystem->setMinimumSize( mColor.checkSystem->sizeHint() );
  connect( mColor.checkSystem, SIGNAL(toggled(bool)),
	   SLOT( slotColorSystem(bool)));
  topLayout->addWidget( mColor.checkSystem, 0, AlignLeft );

  QFrame *hline = new QFrame( page );
  hline->setFrameStyle( QFrame::Sunken | QFrame::HLine );
  hline->setFixedHeight( hline->sizeHint().height() );
  topLayout->addWidget( hline );

  QFrame *frame = new QFrame( page );
  if( frame == 0 ) { return; }
  topLayout->addWidget( frame, 0, AlignLeft );
  QGridLayout *gbox = new QGridLayout( frame, 12, 2, 0, INNER_SEP );
  if( gbox == 0 ) { return; }
 
  KColorButton *button[12];
  QLabel *label[12];
  const char *labeltext[12] = 
  {
    i18n("&Text background"),
    i18n("&Inactive background"),
    i18n("&Even column text"),
    i18n("Odd colum&n text"),
    i18n("O&ffset text"),
    i18n("&Secondary text"),
    i18n("&Marked background"),
    i18n("Mar&ked text"),
    i18n("Cursor te&xt (block shape)"),
    i18n("Cursor &background"),
    i18n("Se&parator"),
    i18n("Non p&rintable text"),
  };

  QColor color[12] = 
  { 
    white, gray, black, blue, red, black, blue, white, black, red, 
    darkGreen, red,
  };

  int w = 0;

  for( uint i=0; i<12; i++ )
  {
    button[i] = new KColorButton( color[i], frame );
    button[i]->setFixedHeight( button[i]->sizeHint().height() );
    if( w<button[i]->sizeHint().width() ) { w=button[i]->sizeHint().width(); }

    label[i] = new QLabel( button[i], labeltext[i], frame );
    label[i]->setFixedHeight( label[i]->sizeHint().height() );
    if( w<label[i]->sizeHint().width() ) { w = label[i]->sizeHint().width(); }
    
    uint col = i%2;
    if( col == 0 )
    {
      gbox->addWidget( label[i], i, 0 );
      gbox->addWidget( button[i], i+1, 0 );
    }
    else
    {
      gbox->addWidget( label[i], i-1, 1 );
      gbox->addWidget( button[i], i, 1 );
    }
  }

  gbox->activate();
  frame->setMinimumWidth( w * 2 + INNER_SEP );

  mColor.textBg      = button[0];
  mColor.inactiveBg  = button[1];
  mColor.evenFg      = button[2];
  mColor.oddFg       = button[3];
  mColor.offsetFg    = button[4];
  mColor.secondaryFg = button[5];
  mColor.markBg      = button[6];
  mColor.markFg      = button[7];
  mColor.cursorFg    = button[8]; 
  mColor.cursorBg    = button[9];
  mColor.sepFg       = button[10];
  mColor.nonPrintFg  = button[11];
  mColor.textLabelBg      = label[0];
  mColor.inactiveLabelBg  = label[1];
  mColor.evenLabelFg      = label[2];
  mColor.oddLabelFg       = label[3];
  mColor.offsetLabelFg    = label[4];
  mColor.secondaryLabelFg = label[5]; 
  mColor.markLabelBg      = label[6];
  mColor.markLabelFg      = label[7];
  mColor.cursorLabelFg    = label[8];
  mColor.cursorLabelBg    = label[9];
  mColor.sepLabelFg       = label[10];
  mColor.nonPrintLabelFg  = label[11];

  topLayout->addStretch(10);
  topLayout->activate();

}




void COptionDialog::SetupFontPage( void )
{
  const char *text;
  QFrame *page = mFrame[ page_font ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );
  if( topLayout == 0 ) { return; }

  text = i18n("&Use system font (as chosen in Control Center)");
  mFont.checkSystem = new QCheckBox( text, page );
  mFont.checkSystem->setFixedHeight( mFont.checkSystem->sizeHint().height());
  connect( mFont.checkSystem, SIGNAL(toggled(bool)),
	   SLOT( slotFontSystem(bool)));
  topLayout->addWidget( mFont.checkSystem );

  QFrame *frame = new QFrame( page );
  if( frame == 0 ) { return; }
  frame->setFrameStyle( QFrame::Box|QFrame::Sunken );
  topLayout->addWidget( frame );

  QBoxLayout *hbox = new QHBoxLayout( frame, 2 );
  if( hbox == 0 ) { return; }
  mFont.chooser = new KFontChooser( frame, "font" );
  if( mFont.chooser == 0 ) { return; }
  hbox->addWidget( mFont.chooser );

  hbox = new QHBoxLayout();
  if( hbox == 0 ) { return; }
  topLayout->addLayout( hbox );
  
  mFont.nonPrintInput = new QLineEdit( page );
  if( mFont.nonPrintInput == 0 ) { return; }
  mFont.nonPrintInput->setFixedHeight(
    mFont.nonPrintInput->sizeHint().height() );
  mFont.nonPrintInput->setMaxLength( 1 );
 
  text = i18n("Map non printable characters to:");
  QLabel *nonPrintLabel = new QLabel( mFont.nonPrintInput, text, page );
  if( nonPrintLabel == 0 ) { return; }
  nonPrintLabel->setFixedHeight( mFont.nonPrintInput->sizeHint().height() );
  nonPrintLabel->setFixedWidth( nonPrintLabel->sizeHint().width() );
  
  hbox->addWidget( nonPrintLabel, 0, AlignLeft );
  hbox->addSpacing( INNER_SEP );
  hbox->addWidget( mFont.nonPrintInput, 10 );


  mFont.chooser->setFont( QFont( "courier", 14, QFont::Bold, FALSE ), true );

  topLayout->addStretch(10);
  topLayout->activate();
}


void COptionDialog::SetupMiscPage( void )
{
  const char *text;
  QFrame *page = mFrame[ page_misc ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );

  QFrame *f1 = new QFrame( page );
  if( f1 == 0 ) { return; }
  topLayout->addWidget( f1 );

  QHBoxLayout *hbox = new QHBoxLayout( f1 );
  if( hbox == 0 ) { return; }

  QFrame *f2 = new QFrame( f1 );
  if( f2 == 0 ) { return; }
  QFrame *f3 = new QFrame( f1 );
  if( f3 == 0 ) { return; }
  hbox->addWidget( f2 );
  hbox->addSpacing( 10 );
  hbox->addWidget( f3 );

  QVBoxLayout *vbox1 = new QVBoxLayout( f2 );
  if( vbox1 == 0 ) { return; }
  QVBoxLayout *vbox2 = new QVBoxLayout( f3 );
  if( vbox2 == 0 ) { return; }

  mMisc.undoSpin = new QSpinBox( f2 );
  mMisc.undoSpin->setFixedHeight( mMisc.undoSpin->sizeHint().height() );
  mMisc.undoSpin->setRange( 10, 10000 );
  mMisc.undoSpin->setSteps( 5, 5 );
  mMisc.undoSpin->setValue( 50 );
  QLabel *label = new QLabel( mMisc.undoSpin, i18n("&Undo history"), f2 );
  label->setFixedSize( label->sizeHint() );
  vbox1->addWidget( label, 0, AlignLeft );
  vbox1->addSpacing( 5 );
  vbox1->addWidget( mMisc.undoSpin, 0, AlignLeft  );

  mMisc.openCombo = new QComboBox( false, f3 );
  mMisc.openCombo->setFixedHeight( mMisc.undoSpin->sizeHint().height() );
  const char *items[] = 
  { 
    i18n( "None" ),
    i18n( "Most recent file" ),
    i18n( "All recent files" ),
    0,
  };
  mMisc.openCombo->insertStrList( items );
  text = i18n("Open files on &startup");
  label = new QLabel( mMisc.openCombo, text, f3 );
  label->setFixedSize( label->sizeHint() );
  vbox2->addWidget( label, 0, AlignLeft );
  vbox2->addSpacing( 5 );
  vbox2->addWidget( mMisc.openCombo, 0, AlignLeft );

  QGroupBox *group = new QGroupBox( page );
  if( group == 0 ) { return; }
  group->setTitle( i18n("Sounds") );
  topLayout->addWidget( group );
  QVBoxLayout *vbox = new QVBoxLayout( group, 5 );
  vbox->addSpacing( group->fontMetrics().height() );
  text = i18n("Make sound on data &input (eg. typing) failure");
  mMisc.inputCheck = new QCheckBox( text, group );
  mMisc.inputCheck->setFixedSize( mMisc.inputCheck->sizeHint() );
  vbox->addWidget( mMisc.inputCheck, 0, AlignLeft );
  text = i18n("Make sound on &fatal failure");
  mMisc.fatalCheck = new QCheckBox( text, group );
  mMisc.fatalCheck->setFixedSize( mMisc.fatalCheck->sizeHint() );
  vbox->addWidget( mMisc.fatalCheck, 0, AlignLeft );
  vbox->activate();

  text = i18n("Auto&matic copy to clipboard when selection is ready");
  mMisc.autoCheck = new QCheckBox( text, page );
  mMisc.autoCheck->setFixedSize( mMisc.autoCheck->sizeHint() );
  topLayout->addWidget( mMisc.autoCheck, 0, AlignLeft );

  text = i18n("&Editor starts in \"insert\" mode" );
  mMisc.insertCheck = new QCheckBox( text, page );
  mMisc.insertCheck->setFixedHeight( mMisc.insertCheck->sizeHint().height() );
  topLayout->addWidget( mMisc.insertCheck, 0, AlignLeft );

  text = i18n("Open file with write &proctection enabled");
  mMisc.writeProtectCheck = new QCheckBox( text, page );
  mMisc.writeProtectCheck->setFixedSize( mMisc.writeProtectCheck->sizeHint() );
  topLayout->addWidget( mMisc.writeProtectCheck, 0, AlignLeft );

  text = i18n("Confirm &wrapping (to beginning or end) during search");
  mMisc.confirmWrapCheck = new QCheckBox( text, page );
  mMisc.confirmWrapCheck->setMinimumSize( mMisc.confirmWrapCheck->sizeHint() );
  topLayout->addWidget( mMisc.confirmWrapCheck, 0, AlignLeft );

  text = i18n("Cursor jumps to &nearest byte when moved");
  mMisc.cursorJumpCheck = new QCheckBox( text, page );
  mMisc.cursorJumpCheck->setFixedSize( mMisc.cursorJumpCheck->sizeHint() );
  topLayout->addWidget( mMisc.cursorJumpCheck, 0, AlignLeft );

  topLayout->addStretch(10);
  topLayout->activate();
}



void COptionDialog::slotModeSelectorChanged( int index )
{
  int lineSize = mDisplayState.line.getLineSize( index );
  int colSize  = mDisplayState.line.getColumnSize( index );
  mLayout.lineSizeSpin->setValue( lineSize );
  mLayout.columnSizeSpin->setValue( colSize );
}

void COptionDialog::slotLineSizeChanged( int value )
{
  int index = mLayout.formatCombo->currentItem();
  mDisplayState.line.setLineSize( index, value );
  if( index == 3 ) { mLayout.columnSizeSpin->setValue( value ); }
}

void COptionDialog::slotColumnSizeChanged( int value )
{
  int index = mLayout.formatCombo->currentItem();
  mDisplayState.line.setColumnSize( index, value );
  if( index == 3 ) { mLayout.lineSizeSpin->setValue( value ); }
}

void COptionDialog::slotColumnSepCheck( bool state )
{
  state = state == true ? false: true;
  mLayout.columnSepLabel->setEnabled( state );
  mLayout.columnSepSpin->setEnabled( state );
}


void COptionDialog::slotBlinkIntervalCheck( bool state )
{
  state = state == true ? false : true;
  mCursor.blinkSpin->setEnabled( state );
  mCursor.blinkLabel->setEnabled( state );
}

void COptionDialog::slotBlockCursorCheck( bool state )
{
  state = state == true ? false : true;
  mCursor.thickCheck->setEnabled( state );
}

void COptionDialog::slotFontSystem( bool state )
{
  mFont.chooser->setEnabled( state == true ? false : true );
}

void COptionDialog::slotColorSystem( bool state )
{
  state = state == true ? false : true;
  mColor.textBg->setEnabled( state );
  mColor.inactiveBg->setEnabled( state );
  mColor.evenFg->setEnabled( state );
  mColor.oddFg->setEnabled( state );
  mColor.offsetFg->setEnabled( state );
  mColor.secondaryFg->setEnabled( state );
  mColor.markBg->setEnabled( state );
  mColor.markFg->setEnabled( state );
  mColor.cursorFg->setEnabled( state );
  mColor.cursorBg->setEnabled( state );
  mColor.sepFg->setEnabled( state );
  mColor.nonPrintFg->setEnabled( state );
  mColor.textLabelBg->setEnabled( state );
  mColor.inactiveLabelBg->setEnabled( state );
  mColor.evenLabelFg->setEnabled( state );
  mColor.oddLabelFg->setEnabled( state );
  mColor.offsetLabelFg->setEnabled( state );
  mColor.secondaryLabelFg->setEnabled( state );
  mColor.markLabelBg->setEnabled( state );
  mColor.markLabelFg->setEnabled( state );
  mColor.cursorLabelFg->setEnabled( state );
  mColor.cursorLabelBg->setEnabled( state );
  mColor.sepLabelFg->setEnabled( state );
  mColor.nonPrintLabelFg->setEnabled( state );
}

void COptionDialog::slotDefault( void )
{
  //
  // We use the constructor settings as default values
  //
  switch( activePage() )
  {
    case page_layout:
    {
      SDisplayLayout layout; 
      SDisplayLine line; 
      setLayout( layout, line );
    }
    break;

    case page_cursor:
    {
      SDisplayCursor cursor;
      setCursor( cursor );
    }
    break;

    case page_color:
    {
      SDisplayColor color;
      setColor( color );
    }
    break;

    case page_font:
    {
      SDisplayFont font;
      setFont( font );
    }
    break;

    case page_misc:
    {
      SDisplayMisc misc;
      setMisc( misc );
    }
    break;
  }
}


void COptionDialog::slotApply( void )
{
  switch( activePage() )
  {
    case page_layout:
    {
      int index = mLayout.formatCombo->currentItem();
      mDisplayState.line.setLineSize(index, mLayout.lineSizeSpin->value());
      mDisplayState.line.setColumnSize(index, mLayout.columnSizeSpin->value());
      emit lineSizeChoise( mDisplayState.line );

      SDisplayLayout &l = mDisplayState.layout;
      l.lockLine = mLayout.lockLineCheck->isChecked();
      l.lockColumn = mLayout.lockColumnCheck->isChecked();
      l.leftSeparatorWidth = mLayout.leftSepWidthSpin->value();
      l.rightSeparatorWidth = mLayout.rightSepWidthSpin->value();
      l.separatorMarginWidth = mLayout.separatorSpin->value();
      l.edgeMarginWidth = mLayout.edgeSpin->value();
      l.leftSeparatorWidth = mLayout.leftSepWidthSpin->value();
      l.rightSeparatorWidth = mLayout.rightSepWidthSpin->value();
      l.columnCharSpace = mLayout.columnCheck->isChecked();
      l.columnSpacing = mLayout.columnSepSpin->value();
      emit layoutChoise( mDisplayState.layout );
    }
    break;


    case page_cursor:
    {
      SDisplayCursor &c = mDisplayState.cursor;      
      c.alwaysVisible    = mCursor.blinkCheck->isChecked();
      c.interval         = mCursor.blinkSpin->value();
      c.alwaysBlockShape = mCursor.blockCheck->isChecked();
      c.thickInsertShape = mCursor.thickCheck->isChecked();
      c.focusMode        = cursorFocusMode();
      emit cursorChoise( mDisplayState.cursor );
    }
    break;

    case page_color:
    {
      SDisplayColor &c = mDisplayState.color;
      c.useSystemColor = mColor.checkSystem->isChecked();
      c.textBg = mColor.textBg->color();
      c.inactiveBg = mColor.inactiveBg->color();
      c.markBg = mColor.markBg->color();
      c.markFg = mColor.markFg->color();
      c.primaryFg[0] = mColor.evenFg->color();
      c.primaryFg[1] = mColor.oddFg->color();
      c.offsetFg = mColor.offsetFg->color();
      c.secondaryFg = mColor.secondaryFg->color();
      c.cursorBg = mColor.cursorBg->color();
      c.cursorFg = mColor.cursorFg->color();
      c.leftSeparatorFg = mColor.sepFg->color();
      c.rightSeparatorFg = mColor.sepFg->color(); 
      c.nonPrintFg = mColor.nonPrintFg->color(); 
      emit colorChoise( mDisplayState.color );
    }
    break;

    case page_font:
    {
      SDisplayFont &f = mDisplayState.font;
      f.useSystemFont = mFont.checkSystem->isChecked();
      f.localFont = mFont.chooser->font();

      const char *nonPrintText = mFont.nonPrintInput->text(); 
      if( nonPrintText == 0 || strlen(nonPrintText) == 0 )
      {
	f.nonPrintChar = ' ';
      }
      else
      {
	f.nonPrintChar = nonPrintText[0];
      }

      emit fontChoise( mDisplayState.font );
    }
    break;

    case page_misc:
    {
      SDisplayMisc &m = mDisplayState.misc;
      m.undoLevel = mMisc.undoSpin->value();
      m.openFile  = (SDisplayMisc::EOpenFile)mMisc.openCombo->currentItem();
      m.autoCopyToClipboard = mMisc.autoCheck->isChecked();
      m.insertMode = mMisc.insertCheck->isChecked();
      m.writeProtect = mMisc.writeProtectCheck->isChecked();
      m.inputSound = mMisc.inputCheck->isChecked();
      m.confirmWrap = mMisc.confirmWrapCheck->isChecked();
      m.cursorJump = mMisc.cursorJumpCheck->isChecked();
      m.fatalSound = mMisc.fatalCheck->isChecked();
      emit miscChoise( mDisplayState.misc );
    }
    break;
  }

}

void COptionDialog::setLayout( SDisplayLayout &layout, SDisplayLine &line )
{
  mDisplayState.line = line;
  mDisplayState.layout = layout;

  slotModeSelectorChanged( mLayout.formatCombo->currentItem() );
  mLayout.lockLineCheck->setChecked( layout.lockLine );
  mLayout.lockColumnCheck->setChecked( layout.lockColumn );
  mLayout.leftSepWidthSpin->setValue( layout.leftSeparatorWidth );
  mLayout.rightSepWidthSpin->setValue( layout.rightSeparatorWidth );
  mLayout.separatorSpin->setValue( layout.separatorMarginWidth );
  mLayout.edgeSpin->setValue( layout.edgeMarginWidth );
  mLayout.leftSepWidthSpin->setValue( layout.leftSeparatorWidth );
  mLayout.rightSepWidthSpin->setValue( layout.rightSeparatorWidth );
  mLayout.columnCheck->setChecked( layout.columnCharSpace ); 
  slotColumnSepCheck( layout.columnCharSpace );
  mLayout.columnSepSpin->setValue( layout.columnSpacing );

}


void COptionDialog::setCursor( SDisplayCursor &cursor )
{
  mDisplayState.cursor = cursor;
 
  mCursor.blinkCheck->setChecked( cursor.alwaysVisible );
  mCursor.blinkSpin->setValue( cursor.interval );
  mCursor.blockCheck->setChecked( cursor.alwaysBlockShape );
  mCursor.thickCheck->setChecked( cursor.thickInsertShape );
  bool state = cursor.focusMode == SDisplayCursor::stopBlinking ? true : false;
  mCursor.stopRadio->setChecked( state );
  state = cursor.focusMode == SDisplayCursor::hide ? true : false;
  mCursor.hideRadio->setChecked( state );
  state = cursor.focusMode == SDisplayCursor::ignore ? true : false;
  mCursor.nothingRadio->setChecked( state );
}


void COptionDialog::setColor( SDisplayColor &color )
{
  mDisplayState.color = color;

  mColor.checkSystem->setChecked( color.useSystemColor );
  mColor.textBg->setColor( color.textBg );
  mColor.inactiveBg->setColor( color.inactiveBg );
  mColor.markBg->setColor( color.markBg );
  mColor.markFg->setColor( color.markFg );
  mColor.evenFg->setColor( color.primaryFg[0] );
  mColor.oddFg->setColor( color.primaryFg[1] );
  mColor.offsetFg->setColor( color.offsetFg );
  mColor.secondaryFg->setColor( color.secondaryFg );
  mColor.sepFg->setColor( color.leftSeparatorFg );
  mColor.nonPrintFg->setColor( color.nonPrintFg ); 
  mColor.cursorBg->setColor( color.cursorBg );
  mColor.cursorFg->setColor( color.cursorFg );
 
}


void COptionDialog::setFont( SDisplayFont &font )
{
  mDisplayState.font = font;

  mFont.checkSystem->setChecked( font.useSystemFont );
  mFont.chooser->setFont( font.localFont, true );
  char buf[2] = { font.nonPrintChar, 0 };
  mFont.nonPrintInput->setText( buf );
}


void COptionDialog::setMisc( SDisplayMisc &misc )
{
  mDisplayState.misc = misc;

  mMisc.undoSpin->setValue( misc.undoLevel );
  mMisc.openCombo->setCurrentItem( (uint)misc.openFile );
  mMisc.autoCheck->setChecked( misc.autoCopyToClipboard );
  mMisc.insertCheck->setChecked( misc.insertMode );
  mMisc.writeProtectCheck->setChecked( misc.writeProtect );
  mMisc.confirmWrapCheck->setChecked( misc.confirmWrap );
  mMisc.cursorJumpCheck->setChecked( misc.cursorJump );
  mMisc.inputCheck->setChecked( misc.inputSound );
  mMisc.fatalCheck->setChecked( misc.fatalSound );
}


void COptionDialog::setState( SDisplayState &state )
{
  setLayout( state.layout, state.line );
  setCursor( state.cursor );
  setColor( state.color );
  setFont( state.font );
  setMisc( state.misc );
}


SDisplayCursor::EFocusMode COptionDialog::cursorFocusMode( void )
{
  if( mCursor.stopRadio->isChecked() == true )
  {
    return( SDisplayCursor::stopBlinking );
  }
  else if( mCursor.hideRadio->isChecked() == true )
  {
    return( SDisplayCursor::hide );
  }
  else
  {
    return( SDisplayCursor::ignore );
  }
}







